
/**************************************************************************
 *
 *  $Id: cmp_time_util.h 1.2 2018/08/08 08:23:28Z martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Definitions and prototypes for cmp_time_util.c.
 *
 * -----------------------------------------------------------------------
 *  $Log: cmp_time_util.h $
 *  Revision 1.2  2018/08/08 08:23:28Z  martin
 *  Moved some functions and structures to more convenient files.
 *  Fixed Windows build.
 *  Updated function prototypes.
 *  Revision 1.1  2013/07/25 10:42:59  martin
 *  Initial revision.
 *
 **************************************************************************/

#ifndef _CMP_TIME_UTIL_H
#define _CMP_TIME_UTIL_H


/* Other headers to be included */

#include <mbgdevio.h>
#include <deviohlp.h>

#ifdef _CMP_TIME_UTIL
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif


/* Start of header body */

#if 0 && defined( _USE_PACK )  // use default alignment
  #pragma pack( 1 )      // set byte alignment
  #define _USING_BYTE_ALIGNMENT
#endif


#ifdef __cplusplus
extern "C" {
#endif

_ext MBG_PC_CYCLES_FREQUENCY cyc_freq;

#if defined( MBG_TGT_WIN32 )
  #define snprintf _snprintf
#endif



/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 /**
 * @brief Check if both devices support fast HR timestamps.
 *
 * @param[in]  dh1          Handle of the first device
 * @param[in]  p_dev_1      Device info of the first device
 * @param[in]  dh2          Handle of the second device
 * @param[in]  p_dev_2      Device info of the second device
 * @param[in]  err_msg_fnc  Pointer to a function to be called in case of error
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int chk_fast_tstamp_supp( MBG_DEV_HANDLE dh1, const PCPS_DEV *p_dev_1, MBG_DEV_HANDLE dh2, const PCPS_DEV *p_dev_2, MBG_ERR_MSG_FNC err_msg_fnc ) ;

 /**
 * @brief Read timestamps and cycles from both devices
 *
 * @param[in]   dh1        Handle of the first device
 * @param[out]  p_htc1     Timestamp and cycles read from the first device
 * @param[in]   dh2        Handle of the second device
 * @param[out]  p_htc2     Timestamp and cycles read from the first device
 * @param[in]   read_fast  Flag indicating that fast timestamps should be read
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int get_htc_timestamps( MBG_DEV_HANDLE dh1, PCPS_HR_TIME_CYCLES *p_htc1, MBG_DEV_HANDLE dh2, PCPS_HR_TIME_CYCLES *p_htc2, int read_fast ) ;

 /**
 * @brief Compute the delta cycles and time difference
 *
 * @param[in]   p_htc        Timestamp/cycles t from device under test.
 * @param[in]   p_htc_ref    Timestamp/cycles t_ref from reference device.
 * @param[out]  p_delta_ts   Pointer to save time difference (t - tref), can be NULL.
 * @param[out]  p_delta_cyc  Pointer to save delta cycles (t - tref), can be NULL.
 *
 * @return The difference between delta_t and delta_cycles.
 */
 double get_htc_delta( const PCPS_HR_TIME_CYCLES *p_htc, const PCPS_HR_TIME_CYCLES *p_htc_ref, double *p_delta_ts, double *p_delta_cyc ) ;

 /**
 * @brief Print UTC date and time from a ::PCPS_TIME_STAMP structure to a string
 *
 * This function is mostly similar to ::mbg_snprint_hr_tstamp, but optionally
 * appends the raw timestamp instead of inserting it at the beginning.
 *
 * @param[out]  s          Address of a string buffer to be filled.
 * @param[in]   max_len    Size of the string buffer.
 * @param[in]   p          Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 * @param[in]   print_raw  Flag indicating if a raw timestamp (hex) is to be appended.
 *
 * @return The number of characters printed to the buffer.
 *
 * @see ::mbg_snprint_hr_tstamp
 * @see ::mbg_snprint_hr_time_loc
 */
 int mbg_snprint_hr_tstamp_ext( char *s, int max_len, const PCPS_TIME_STAMP *p, int print_raw ) ;

 /**
 * @brief Print local date and time from a ::PCPS_TIME_STAMP structure to a string
 *
 * This function is mostly similar to ::mbg_snprint_hr_tstamp, but optionally
 * appends the raw timestamp instead of inserting it at the beginning.
 *
 * @param[out]  s          Address of a string buffer to be filled.
 * @param[in]   max_len    Size of the string buffer.
 * @param[in]   p          Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 *
 * @return The number of characters printed to the buffer.
 *
 * @see ::mbg_snprint_hr_tstamp
 * @see ::mbg_snprint_hr_tstamp_ext
 */
 int mbg_snprint_hr_time_loc( char *s, size_t max_len, const PCPS_HR_TIME *p ) ;


/* ----- function prototypes end ----- */

#ifdef __cplusplus
}
#endif


#if defined( _USING_BYTE_ALIGNMENT )
  #pragma pack()      // set default alignment
  #undef _USING_BYTE_ALIGNMENT
#endif

/* End of header body */


#undef _ext
#undef _DO_INIT

#endif  /* _CMP_TIME_UTIL_H */

