
/**************************************************************************
 *
 *  $Id: mbgmktm.c 1.2 2017/07/05 10:00:29Z martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Function to convert broken down time to Unix time (seconds since 1970)
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgmktm.c $
 *  Revision 1.2  2017/07/05 10:00:29Z  martin
 *  Let mbg_mktime() fail if year is out of a range which depends on
 *  the size of the 'time_t' type provided by the build environment.
 *  Allow 0 as valid return value for mbg_mktime().
 *  Added some doxygen comments.
 *  Revision 1.1  2006/08/22 08:57:15  martin
 *  Former function totalsec() moved here from pcpsmktm.c.
 *
 **************************************************************************/

#define _MBGMKTM
 #include <mbgmktm.h>
#undef _MBGMKTM

#include <assert.h>


static const char Days[12] =
{
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

static int YDays[12] =
{
  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334
};



/*HDR*/
/**
 * @brief Compute a linear time_t value from broken down date and time
 *
 * This function works like the standard mktime() function but does not
 * account for a timezone setting configured for the standard C library.
 * Also, it does not take a structure but a set of variables which makes
 * it more versatile. The accepted variables are in the same ranges
 * as the struct tm members used by mktime().
 *
 * @param[in] year   year number - 1900
 * @param[in] month  months since January, 0..11
 * @param[in] day    days after 1st, 0..30
 * @param[in] hour   0..23
 * @param[in] min    0..59
 * @param[in] sec    0..59, 60 if leap second
 *
 * @return seconds since 1970-01-01 (Unix time_t format) or ((time_t) -1) if range overflow
 */
time_t mbg_mktime( int year, int month, int day,
                   int hour, int min, int sec )
{
  int leaps;
  long days;
  time_t secs;

  // time_t should be at least 4 bytes
  assert( sizeof( time_t ) >= 4 );

  if ( sizeof( time_t ) == 4 )
  {
    // 32 bit *signed* time_t will roll over after 2038-01-19 03:14:07
    // when the number of seconds reaches 0x7FFFFFFF, so we don't try
    // conversions for 2038 or later, though 32 bit *unsigned* time_t
    // may still work after year 2100.
    if ( year < 70 || year > 137 )
      goto fail;
  }
  else
    if ( sizeof( time_t ) == 8 )
    {
      // 64 bit time_t will work for million years. However, it's not
      // clear what happes for dates before 1970-01-01T00:00:00 if time_t
      // is *unsigned*.
      if ( year < 70 )
        goto fail;
    }
    else
      goto fail;

  min += sec / 60;
  sec %= 60;              /* Seconds are normalized */
  hour += min / 60;
  min %= 60;              /* Minutes are normalized */
  day += hour / 24;
  hour %= 24;             /* Hours are normalized   */

  year += month / 12;     /* Normalize month (not necessarily final) */
  month %= 12;

  while ( day >= Days[month] )
  {
    if ( !( year & 3 ) && ( month == 1 ) )
    {
      if (day > 28)
      {
        day -= 29;
        month++;
      }
      else
        break;
    }
    else
    {
      day -= Days[month];
      month++;
    }

    year += month / 12; /* Normalize month */
    month %= 12;
  }

  year -= 70;
  leaps = ( year + 2 ) / 4;

  if ( !( ( year + 70 ) & 3 ) && ( month < 2 ) )
    --leaps;

  days = year * 365L + leaps + YDays[month] + day;

  secs = days * 86400L + hour * 3600L + min * 60L + sec;

  if ( secs < 0 )  // == 0 is valid for 1970-01-01 00:00:00
    goto fail;

  return secs;


fail:
  return (time_t) -1;

}  // mbg_mktime

