
/**************************************************************************
 *
 *  $Id: toolutil.h 1.8.1.1 2019/07/22 09:52:37Z martin TRASH $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Definitions and prototypes for toolutil.c.
 *
 * -----------------------------------------------------------------------
 *  $Log: toolutil.h $
 *  Revision 1.8.1.1  2019/07/22 09:52:37Z  martin
 *  mbg_print_opt_info() now accepts printf-like format string as
 *  2nd argument, plus associated optional parameters.
 *  Revision 1.8  2019/03/13 09:44:57  martin
 *  Moved predefined program exit codes to mbgerror.h.
 *  Revision 1.7  2018/12/14 13:14:52  martin
 *  Updated function prototypes.
 *  Revision 1.6  2018/11/15 12:21:10  martin
 *  Updated function prototypes.
 *  Revision 1.5  2018/06/25 13:47:45  martin
 *  Updated function prototypes.
 *  Revision 1.4  2017/07/05 07:38:06  martin
 *  Support Windows target.
 *  Defined function type MBG_DEV_HANDLER_FNC.
 *  Defined some OS-specific strings e.g. for example
 *  file and device names used in help messages.
 *  Check for MBG_TGT_POSIX instead of MBG_TGT_UNIX.
 *  Doxygen comments.
 *  Updated function prototypes.
 *  Revision 1.3  2012/10/15 09:36:22  martin
 *  Use common way to handle version information.
 *  Added string table with PZF state names.
 *  Updated function prototypes.
 *  Revision 1.2  2009/06/19 12:11:35  martin
 *  Updated function prototypes.
 *  Revision 1.1  2008/12/17 10:45:14  martin
 *  Initial revision.
 *  Revision 1.1  2008/12/15 08:35:08  martin
 *  Initial revision.
 *
 **************************************************************************/

#ifndef _TOOLUTIL_H
#define _TOOLUTIL_H


/* Other headers to be included */

#include <mbgdevio.h>
#include <mbgversion.h>

#if defined( MBG_TGT_POSIX )

  #include <unistd.h>

#elif defined( MBG_TGT_WIN32 )

  #include <io.h>
  #include <fcntl.h>
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <wingetopt.h>

  #define open          _open
  #define snprintf      _snprintf
  #define sleep( _x )   Sleep( (_x) * 1000 )
  #define usleep( _x )  Sleep( (_x) / 1000 )

#endif



#ifdef _TOOLUTIL
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif


/* Start of header body */

#ifdef __cplusplus
extern "C" {
#endif


enum CHK_DEV_FLAGS
{
  CHK_DEV_ALL_DEVICES = 0x0001,   ///< call callback for all devices
  CHK_DEV_WITHOUT_DEV = 0x0002    ///< call callback once if no device found
};



#if MBG_TGT_HAS_DEV_FN
  #define EXAMPLE_DEV_FN_1      MBGCLOCK_DEV_FN_BASE "0"
  #define EXAMPLE_DEV_FN_1_TCR  EXAMPLE_DEV_FN_1
  #define EXAMPLE_DEV_FN_2      MBGCLOCK_DEV_FN_BASE "3"
#endif

#define EXAMPLE_DEV_NAME_1      "gps180pex"
#define EXAMPLE_DEV_NAME_1_TCR  "tcr167pci"
#define EXAMPLE_DEV_NAME_2      "tcr170pex_027911002000"


#if defined( MBG_TGT_POSIX )
  #define ROOT_PRIVILEGES_STR        "with root privileges"
#else
  #define ROOT_PRIVILEGES_STR        "as administrator"
#endif


#if !defined( MBG_EXP_YEAR_LIMIT )
  #define MBG_EXP_YEAR_LIMIT   1980
#endif


_ext uint16_t mbg_exp_year_limit
#ifdef _DO_INIT
 = MBG_EXP_YEAR_LIMIT
#endif
;

_ext int must_print_usage;

_ext const char str_empty[]
#ifdef _DO_INIT
 = ""
#endif
;


_ext const char *pzf_corr_state_name[N_PZF_CORR_STATE]
#ifdef _DO_INIT
 = PZF_CORR_STATE_NAMES_ENG
#endif
;



/**
 * @brief The type of functions to called to handle a device in a specific way.
 */
typedef int MBG_DEV_HANDLER_FNC( MBG_DEV_HANDLE, const PCPS_DEV *);



/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 /**
 * @brief Print the program version to a string buffer
 *
 * @param[out]  s              Address of a string buffer to be filled.
 * @param[in]   max_len        Size of the string buffer.
 *
 * @return The number of characters printed to the buffer.
 */
 int mbg_program_version_str( char *s, size_t max_len ) ;

 /**
 * @brief Print some program info to a string buffer
 *
 * @param[out]  s              Address of a string buffer to be filled.
 * @param[in]   max_len        Size of the string buffer.
 * @param[in]   pname          The program name.
 * @param[in]   first_year     First copyright year.
 * @param[in]   last_year      Last copyright year.
 *
 * @return The number of characters printed to the buffer.
 */
 int mbg_program_info_str( char *s, size_t max_len, const char *pname, int first_year, int last_year ) ;

 /**
 * @brief Print program info to console
 *
 * @param[in]   pname          The program name.
 * @param[in]   first_year     First copyright year.
 * @param[in]   last_year      Last copyright year.
 */
 void mbg_print_program_info( const char *pname, int first_year, int last_year ) ;

 /**
 * @brief Print usage intro to console
 *
 * @param[in]   pname  The program name.
 * @param[in]   info   An optional additional info string, may be NULL.
 */
 void mbg_print_usage_intro( const char *pname, const char *info ) ;

 /**
 * @brief Print info on a single program option / argument
 *
 * @param[in]   opt_name  The option name, optional, may be NULL.
 * @param[in]   opt_info  The option info, optional, may be NULL.
 */
 __attribute__( ( format( printf, 2, 3 ) ) ) void mbg_print_opt_info( const char *opt_name, const char *opt_info_fmt, ... ) ;

 /**
 * @brief Print info on common program help arguments
 *
 * Lists program parameters causing printing
 * of help / usage information.
 */
 void mbg_print_help_options( void ) ;

 /**
 * @brief Print common info on how to specify devices on the command line
 */
 void mbg_print_device_options( void ) ;

 /**
 * @brief Print program info and default usage information
 *
 * @param[in]   pname      The program name.
 * @param[in]   prog_info  An optional additional info string, may be NULL.
 */
 void mbg_print_default_usage( const char *pname, const char *prog_info ) ;

 /**
 * @brief Retrieve and print some common device info
 *
 * @param[in]   dh        Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[in]   dev_name  A device name string to be printed
 * @param[out]  p_dev     Pointer to a device info structure to be read from the device
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int mbg_get_show_dev_info( MBG_DEV_HANDLE dh, const char *dev_name, PCPS_DEV *p_dev ) ;

 /**
 * @brief Print device info and take some action on a specific device
 *
 * @param[in]   dh        Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[in]   dev_name  A device name string to be printed.
 * @param[in]   fnc       Pointer to a callback function that actually takes the action.
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int mbg_handle_device( MBG_DEV_HANDLE dh, const char *dev_name, MBG_DEV_HANDLER_FNC *fnc ) ;

 /**
 * @brief Get the number of devices actually present
 *
 * Do a real search only when called for the first time.
 *
 * @return The number of devices currently present.
 */
 int chk_get_num_devices( void ) ;

 /**
 * @brief Try to open a device using a parameter string or device index.
 *
 * @param[out]  p_dh           Address of a device handle variable to be set on success.
 * @param[in]   dev_param_str  An optional device specification string. This can be:
 *                             - A device file name, if the operating system supports this. See ::MBG_DEV_FN.
 *                             - A device model name like "GPS180PEX", eventually with the serial number
 *                               appended after an underscore, as in "GPS180PEX_029511026220".
 *                               See ::MBG_DEV_NAME.
 *                             - A device index number as string, e.g "0" or "3".
 *                             - NULL, in which case @p dev_idx has to be a valid device index number.
 * @param[in]   dev_idx               A device index number which is used if @p dev_param_str is NULL.
 * @param[out]  dev_name_buffer       A string buffer to which a device name can be written, see ::MBG_DEV_FN.
 * @param[in]   dev_name_buffer_size  The size of the @p dev_name_buffer buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @see ::mbg_open_device_by_param
 */
 int mbg_open_device_by_param( MBG_DEV_HANDLE *p_dh, const char *dev_param_str, int dev_idx, char *dev_name_buffer, size_t dev_name_buffer_size ) ;

 /**
 * @brief Try to open a device and print a message in case of error.
 *
 * Call ::mbg_open_device_by_param to try to open a device,
 * and print an error message to stderr if the call fails.
 *
 * @param[out]  p_dh           Address of a device handle variable to be set on success.
 * @param[in]   dev_param_str  An optional device specification string. This can be:
 *                             - A device file name, if the operating system supports this. See ::MBG_DEV_FN.
 *                             - A device model name like "GPS180PEX", eventually with the serial number
 *                               appended after an underscore, as in "GPS180PEX_029511026220".
 *                               See ::MBG_DEV_NAME.
 *                             - A device index number as string, e.g "0" or "3".
 *                             - NULL, in which case @p dev_idx has to be a valid device index number.
 * @param[in]   dev_idx               A device index number which is used if @p dev_param_str is NULL.
 * @param[out]  dev_name_buffer       A string buffer to which a device name can be written, see ::MBG_DEV_FN.
 * @param[in]   dev_name_buffer_size  The size of the @p dev_name_buffer buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @see ::mbg_open_device_by_param
 */
 int mbg_open_device_by_param_chk( MBG_DEV_HANDLE *p_dh, const char *dev_param_str, int dev_idx, char *dev_name_buffer, size_t dev_name_buffer_size ) ;

 /**
 * @brief Main action handler that can be called by utility programs
 *
 * This function checks the command line parameters passed to the program.
 * Those which have not been evaluated before are interpreted as device specifiers.
 *
 * Device specifiers can be device file names like "/dev/mbgclock0" on target
 * platforms that support such device names, see ::MBG_DEV_FN,
 * or device type names like "GPS180PEX" which can optionally have a serial
 * number appended, like "GPS180PEX_029511026220", to be able to distinguish
 * between several devices of the same type (see ::MBG_DEV_NAME),
 * or just an index number as required for the ::mbg_open_device call.
 *
 * For each of the specified devices the callback function @p fnc is called to
 * take some specific action on the device.
 *
 * If no device has been specified in the argument list then ::mbg_find_devices
 * is called to set up a device list, and depending on whether ::CHK_DEV_ALL_DEVICES
 * is set in @p chk_dev_flags the callback function @p fnc is either called
 * for every device from the list, or only for the first one.
 *
 * @param[in]   argc     Number of parameters of the program argument list.
 * @param[in]   argv     Array of program argument strings.
 * @param[in]   optind   Number of the command line arguments that have already been handled.
 * @param[in]   fnc      Pointer to a callback function that actually takes an action on each specified device.
 * @param[in]   chk_dev_flags    Bit mask controlling the behavior of the function, see ::CHK_DEV_FLAGS
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int mbg_handle_devices( int argc, char *argv[], int optind, MBG_DEV_HANDLER_FNC *fnc, int chk_dev_flags ) ;

 /**
 * @brief Print date and time from a ::PCPS_TIME structure to a string
 *
 * @param[out]  s        Address of a string buffer to be filled.
 * @param[in]   max_len  Size of the string buffer.
 * @param[in]   p        Pointer to a ::PCPS_TIME structure to be evaluated.
 * @param[in]   verbose  Increase verbosity of the output:<br>
 *                       > 0: append UTC offset and status<br>
 *                       > 1: append signal value
 *
 * @return The number of characters printed to the buffer.
 */
 int mbg_snprint_date_time( char *s, size_t max_len, const PCPS_TIME *p, int verbose ) ;

 /**
 * @brief Print date and time from a ::PCPS_TIME_STAMP structure to a string
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 * @param[in]   utc_offs  A local time offset added to the time stamp before converted to date and time.
 * @param[in]   show_raw  Flag indicating if a raw timestamp (hex) is to be printed, too.
 *
 * @return The number of characters printed to the buffer.
 */
 int mbg_snprint_hr_tstamp( char *s, size_t max_len, const PCPS_TIME_STAMP *p, long utc_offs, int show_raw ) ;

 /**
 * @brief Print date and time from a ::PCPS_HR_TIME structure to a string
 *
 * Converts the UTC timestamp first to local time according to
 * the ::PCPS_HR_TIME::utc_offs value.
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::PCPS_HR_TIME structure to be evaluated.
 * @param[in]   show_raw  Flag indicating if a raw timestamp (hex) is to be printed, too.
 *
 * @return The number of characters printed to the buffer.
 */
 int mbg_snprint_hr_time( char *s, size_t max_len, const PCPS_HR_TIME *p, int show_raw ) ;

 /**
 * @brief Print date and time from a ::PCPS_TIME_STAMP structure
 *
 * First the HR timestamp passed by parameter @p p_ts is printed.
 *
 * If the parameter @p p_prv_ts is not NULL then it should specify
 * an earlier timestamp, and the elapsed time since @p p_ts
 * is appended.
 *
 * Finally the latency value is printed in microseconds, unless
 * parameter @p no_latency is != 0.
 *
 * @param[in]   p_ts         Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 * @param[in]   hns_latency  A latency number in hectonanoseconds, i.e. 100 ns units.
 * @param[in]   p_prv_ts     Pointer to a ::PCPS_TIME_STAMP structure to be evaluated, may be NULL
 * @param[in]   no_latency   A flag indicating if printing the latency should be suppressed.
 * @param[in]   show_raw     Flag indicating if a raw timestamp (hex) is to be printed, too.
 *
 * @see ::mbg_print_hr_time
 */
 void mbg_print_hr_timestamp( PCPS_TIME_STAMP *p_ts, int32_t hns_latency, PCPS_TIME_STAMP *p_prv_ts, int no_latency, int show_raw ) ;

 /**
 * @brief Print date and time from a ::PCPS_HR_TIME structure
 *
 * First the HR timestamp passed by parameter @p p_ht is printed.
 *
 * If the parameter @p p_prv_ts is not NULL then it should specify
 * an earlier timestamp, and the elapsed time since @p p_ht
 * is appended.
 *
 * Next the latency value is printed in microseconds, unless
 * parameter @p no_latency is != 0.
 *
 * @param[in]   p_ht         Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 * @param[in]   hns_latency  A latency number in hectonanoseconds, i.e. 100 ns units.
 * @param[in]   p_prv_ts     Pointer to a ::PCPS_TIME_STAMP structure to be evaluated, may be NULL
 * @param[in]   no_latency   A flag indicating if printing the latency should be suppressed.
 * @param[in]   show_raw     Flag indicating if a raw timestamp (hex) is to be printed, too.
 * @param[in]   verbose      Increase verbosity of the output:<br>
 *                           > 0: append status code<br>
 *                           > 1: append signal value
 *
 * @see ::mbg_print_hr_timestamp
 */
 void mbg_print_hr_time( PCPS_HR_TIME *p_ht, int32_t hns_latency, PCPS_TIME_STAMP *p_prv_ts, int no_latency, int show_raw, int verbose ) ;

 /**
 * @brief Retrieve and print PZF correlation info for a device which supports this
 *
 * @param[in]   dh              Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[in]   show_corr_step  A flag indicating if correlation step indicators are to be appended
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int mbg_show_pzf_corr_info( MBG_DEV_HANDLE dh, int show_corr_step ) ;

 /**
 * @brief Retrieve a ::MBG_GNSS_MODE_INFO structure from a device
 *
 * @param[in]   dh  Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[out]  p   Pointer to a ::MBG_GNSS_MODE_INFO structure to be filled up.
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int mbg_get_gps_gnss_mode_info_chk( MBG_DEV_HANDLE dh, MBG_GNSS_MODE_INFO *p ) ;


/* ----- function prototypes end ----- */


#ifdef __cplusplus
}
#endif


/* End of header body */

#undef _ext
#undef _DO_INIT

#endif  /* _TOOLUTIL_H */

