
/**************************************************************************
 *
 *  $Id: mbg_w32.h 1.10.1.1 2019/01/04 15:39:05Z martin TEST martin.burnicki $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    OS dependend definitions/redefinitions for Win32.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbg_w32.h $
 *  Revision 1.10.1.1  2019/01/04 15:39:05Z  martin
 *  Refactored precise kernel time API usage.
 *  Revision 1.10  2018/12/04 09:52:48Z  martin
 *  Make sure DEBUG_RSRC is always defined.
 *  Revision 1.9  2018/11/22 10:53:18Z  martin
 *  Removed obsolete Windows driver stuff.
 *  Revision 1.8  2018/09/21 10:59:26Z  martin
 *  Renamed PCPS_DDEV field from irp to p_irp.
 *  Use common message functions.
 *  Removed obsolete code.
 *  Revision 1.7  2012/05/30 13:28:43Z  martin
 *  Cleanup.
 *  Revision 1.6  2012/05/29 14:39:53Z  martin
 *  Runtime support for precise time API introduced with Windows 8.
 *  Revision 1.5  2009/03/26 08:32:26Z  martin
 *  Added orgDeviceExtension field for non-PNP driver.
 *  Revision 1.4  2009/01/13 14:45:17Z  martin
 *  Added PCPS_DDEV::RegistryPath field for non-PnP driver.
 *  Use PCPS_DDEV::irp and PCPS_DDEV::win32DeviceName also for non-PnP driver.
 *  Added PCPS_DDEV::DriverObject field.
 *  Conditionally added PCPS_DDEV fields required to support programmable IRQs.
 *  Revision 1.3  2007/09/27 10:36:25Z  martin
 *  Added macro specifying Windows specific device data.
 *  Added USB support.
 *  Renamed Removed to SurpriseRemoved.
 *  Revision 1.2  2002/01/24 09:48:05Z  Udo
 *  check in as lib module
 *  Revision 1.1  2001/09/13 07:12:14Z  Udo
 *  Initial revision
 *
 **************************************************************************/

#ifndef _MBG_W32_H
#define _MBG_W32_H


/* Other headers to be included */

#include <mbg_tgt.h>
#include <words.h>

#if defined( MBG_TGT_WIN32_PNP ) && defined( MBG_TGT_KERNEL )
  #include <usb100.h>
  #include <usbdi.h>
  #include <usbdlib.h>
#endif


#ifdef _MBG_W32
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif


/* Start of header body */

#if !defined( DEBUG_RSRC )
  #define DEBUG_RSRC  0
#endif


#if USE_PGMB_IRQ

  typedef struct
  {
    struct _KINTERRUPT *Object;
    KEVENT *Event;
    KSPIN_LOCK Spinlock;
    KDPC *Dpc;
  } MBG_IRQ_INFO;

  #define _pgmb_irq_vars \
    MBG_IRQ_INFO Irq;

#else

  #define _pgmb_irq_vars

#endif



#if defined( MBG_TGT_KERNEL )

#if !defined( NTKERNELAPI )
  #define NTKERNELAPI
#endif

#if !defined( __in )
  #define __in
#endif

#if !defined( _Out_ )
  #define _Out_
#endif



// definitions for clock() support in kernel drivers


#define clock  w32_clock

// KeQueryTimeIncrement() returns number of 100nsec increments per tick.
// There are 10000000 100nsec increments per second.
#define MBG_TICKS_PER_SEC  ( 10000000 / KeQueryTimeIncrement() )



/**
 * @brief A pointer to a function returning the system time as LARGE_INTEGER.
 *
 * This is for kernel mode only. The function can be e.g. the standard Windows
 * API call KeQuerySystemTime() or the KeQuerySystemTimePrecise() API
 * call introduced with Windows 8.
 */

typedef VOID KE_QUERY_SYSTEM_TIME_FNC( _Out_ PLARGE_INTEGER );


enum KE_QUERY_SYSTEM_TIME_TYPES
{
  KE_QUERY_SYSTEM_TIME_TYPE_WRAPPER_FNC,
  KE_QUERY_SYSTEM_TIME_TYPE_NATIVE_FNC,
  KE_QUERY_SYSTEM_TIME_TYPE_PRECISE_FNC,
  N_KE_QUERY_SYSTEM_TIME_TYPE
};


/**
 * @brief Get precise Windows system time as LARGE_INTEGER.
 *
 * @note This function has been introduced with Windows 8,
 * so it is not available on older Windows versions. Thus
 * the API call is imported from ntoskrnl.exe at runtime.
 *
 * @param[out]  CurrentTime  Address of a LARGE_INTEGER to be filled.
 *
 * @see ::KE_QUERY_SYSTEM_TIME_FNC
 * @see ::import_kernel_precise_time_api
 */
NTKERNELAPI
VOID
KeQuerySystemTimePrecise( _Out_ PLARGE_INTEGER  CurrentTime);

NTKERNELAPI KE_QUERY_SYSTEM_TIME_FNC KeQuerySystemTimePrecise;



#if defined( MBG_TGT_WIN32_PNP )

  #if !defined ( SE_CREATE_SYMBOLIC_LINK_PRIVILEGE )

    // Define some KDD function types which may not be
    // defined in older DDK headers.

    typedef
    NTSTATUS
    IO_COMPLETION_ROUTINE (
        PDEVICE_OBJECT DeviceObject,
        PIRP Irp,
        PVOID Context
        );

    typedef IO_COMPLETION_ROUTINE *PIO_COMPLETION_ROUTINE;


    typedef
    NTSTATUS
    DRIVER_DISPATCH (
        __in struct _DEVICE_OBJECT *DeviceObject,
        __in struct _IRP *Irp
        );

    typedef DRIVER_DISPATCH *PDRIVER_DISPATCH;


    typedef
    VOID
    DRIVER_UNLOAD (
        __in struct _DRIVER_OBJECT *DriverObject
        );

    typedef DRIVER_UNLOAD *PDRIVER_UNLOAD;


    typedef
    NTSTATUS
    DRIVER_ADD_DEVICE (
        __in struct _DRIVER_OBJECT *DriverObject,
        __in struct _DEVICE_OBJECT *PhysicalDeviceObject
        );

    typedef DRIVER_ADD_DEVICE *PDRIVER_ADD_DEVICE;

  #endif  // !defined ( SE_CREATE_SYMBOLIC_LINK_PRIVILEGE )



  #define MBG_USB_MAX_NUMBER_OF_ENDPOINTS      32
  #define MBG_USB_MAX_NUMBER_OF_INTERFACES     1
  #define MBG_USB_DEFAULT_TIMEOUT              5000
  #define MBG_USB_MAX_CONTROL_TRANSFER_TIMEOUT 5000


  #pragma pack( 4 )  //##++++++ this should be obsolete

  typedef struct
  {
    long usage_count;
    int remove_pending;
    KEVENT event;
  } USB_REMOVE_LOCK;


  typedef struct
  {
    int address;
    USBD_PIPE_HANDLE handle;
  } USB_ENDPOINT;


  typedef struct
  {
    int valid;
    int claimed;
    USB_ENDPOINT endpoints[MBG_USB_MAX_NUMBER_OF_ENDPOINTS];
  } USB_INTERFACE;

  #pragma pack()



  #define _pcps_ddev_data_win_pnp           \
    DEVICE_OBJECT *physical_device_object;  \
    USB_REMOVE_LOCK remove_lock;            \
                                            \
    struct                                  \
    {                                       \
      USBD_CONFIGURATION_HANDLE handle;     \
      int value;                            \
      USB_INTERFACE interfaces[MBG_USB_MAX_NUMBER_OF_INTERFACES]; \
    } config;                               \
                                            \
    LONG ref_count;                         \
    POWER_STATE power_state;                \
    DEVICE_POWER_STATE device_power_states[PowerSystemMaximum]; \
                                            \
    DEVICE_OBJECT *NextLowerDriver;         \
    BOOLEAN Started;                        \
    BOOLEAN SurpriseRemoved;                \
    IO_REMOVE_LOCK RemoveLock;              \
    BOOLEAN PortWasMapped;

#else  // if !defined( MBG_TGT_WIN32_PNP )

  // Don't need PNP extensions in non-PNP environment.
  #define _pcps_ddev_data_win_pnp           \
    UNICODE_STRING *RegistryPath;           \
    PVOID orgDeviceExtension;

#endif  // !defined( MBG_TGT_WIN32_PNP )


  #define _pcps_ddev_data_win       \
    DRIVER_OBJECT *DriverObject;    \
    DEVICE_OBJECT *DeviceObject;    \
    UNICODE_STRING win32DeviceName; \
    IRP *p_irp;                     \
    _pgmb_irq_vars                  \
    _pcps_ddev_data_win_pnp

#endif  // defined( MBG_TGT_KERNEL )



#ifdef __cplusplus
extern "C" {
#endif

/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 clock_t w32_clock( void ) ;
 VOID mbg_ke_query_system_time( _Out_ LARGE_INTEGER *p ) ;
 /**
 * @brief Set up a pointer to a function which reads the Windows system time.
 *
 * Starting with Windows 8, the Windows kernel provides a function
 * ::KeQuerySystemTimePrecise which allows reading the system time
 * much more precisely than the original function ::KeQuerySystemTime
 * which is even supported on older Windows versions.
 *
 * This function checks at runtime if the new API call is supported,
 * or not, and a global function pointer ::ke_query_system_time_fnc
 * is set up accordingly with the address of either function.
 * Regardless whether the precise API call is supported or not,
 * the function pointer can be used afterwards whenever a driver
 * needs to read the current system time.
 *
 * This function is for kernel mode only.
 *
 * @see ::ke_query_system_time_fnc
 * @see ::ke_query_system_time_type
 * @see ::KeQuerySystemTimePrecise
 */
 void import_kernel_precise_time_api( void ) ;


/* ----- function prototypes end ----- */


_ext KE_QUERY_SYSTEM_TIME_FNC *ke_query_system_time_fnc
#ifdef _DO_INIT
  = mbg_ke_query_system_time
#endif
;


_ext int ke_query_system_time_type
#ifdef _DO_INIT
  = KE_QUERY_SYSTEM_TIME_TYPE_WRAPPER_FNC
#endif
;  // see ::KE_QUERY_SYSTEM_TIME_TYPES


#ifdef __cplusplus
}
#endif

/* End of header body */

#undef _ext
#undef _DO_INIT

#endif  /* _MBG_W32_H */

