
/**************************************************************************
 *
 *  $Id: timecnv.h 1.15 2018/09/21 08:30:57Z martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Definitions and prototypes for timecnv.c.
 *
 * -----------------------------------------------------------------------
 *  $Log: timecnv.h $
 *  Revision 1.15  2018/09/21 08:30:57Z  martin
 *  Changed year_limit from 1990 to 1970.
 *  Revision 1.14  2018/08/08 09:49:20Z  martin
 *  Support non-Windows builds.
 *  Moved some definitions to more appropriate headers.
 *  New Windows-specific inline functions
 *  mbg_pcps_time_stamp_to_filetime() and
 *  mbg_pcps_tstamp64_to_filetime().
 *  Updated function prototypes.
 *  Revision 1.13  2009/03/17 09:27:08Z  daniel
 *  Updated function prototypes.
 *  Revision 1.12  2009/01/12 13:57:36Z  daniel
 *  Uses ULL instead of LLU.
 *  FILETIME_1970 with 64 bit suffix is different for GnuC and VC.
 *  Added perf_count_on_time in SERIAL_TIME_STRING.
 *  Revision 1.11  2006/05/04 14:13:38Z  martin
 *  Added initializers for time string format names.
 *  Added macro _secs_to_hns().
 *  Revision 1.10  2006/05/02 10:39:09Z  martin
 *  Moved some definitions here and renamed some symbols.
 *  Added new definitions for unified parsing of time strings.
 *  Cleanup and renamed field in SERIAL_TIME_STRING.
 *  Updated function prototypes.
 *  Revision 1.9  2005/11/15 14:36:33Z  martin
 *  Use new type PCPS_TIME_STATUS_X in REF_TIME.
 *  Updated function prototypes.
 *  Revision 1.8  2005/06/09 10:35:18Z  martin
 *  Moved declaration of serclock_cfg elsewhere.
 *  Revision 1.7  2004/04/27 07:41:47Z  martin
 *  Moved some macros here.
 *  Moved some constants from other files here.
 *  Pack structures 1 byte aligned.
 *  Updated function prototypes.
 *  Revision 1.6  2004/02/12 10:16:21Z  martin
 *  Updated function prototypes.
 *  Revision 1.5  2004/01/06 16:26:56Z  martin
 *  Updated function prototypes.
 *  Revision 1.4  2003/12/19 16:13:10Z  martin
 *  New parameter for GET_REF_TIME_FNC type.
 *  Revision 1.3  2003/09/12 12:47:41Z  Daniel
 *  added include file mbgsvctl.h
 *  Revision 1.2  2002/09/30 12:17:12  martin
 *  Added structure SERIAL_TIME_STRING.
 *  Added type of function to get REF_TIME.
 *  Updated function prototypes.
 *  Revision 1.1  2002/08/28 10:47:36Z  Daniel
 *  Initial revision
 *
 **************************************************************************/

#ifndef _TIMECNV_H
#define _TIMECNV_H

/* Other headers to be included */

#include <mbg_tgt.h>
#include <gpsdefs.h>
#include <pcpsdefs.h>
#include <pcpsutil.h>
#include <mbgtime.h>
#include <mbgpccyc.h>

#if defined( MBG_TGT_WIN32 )
  #include <mbgsvcio.h>
  #include <mbgdevio.h>
  #include <mbg_serclock_cfg.h>
  #include <rs232.h>
#endif //MBG_TGT_WIN32

#include <use_pack.h>


#ifdef _TIMECNV
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif


/* Start of header body */

#if defined( _USE_PACK )
  #pragma pack( 1 )      // set byte alignment
  #define _USING_BYTE_ALIGNMENT
#endif


#ifdef __cplusplus
extern "C" {
#endif

// ASCII definitions

#define ASCII_STX       0x02
#define ASCII_ETX       0x03
#define ASCII_BEL       0x07
#define ASCII_BS        0x08
#define ASCII_LF        0x0A
#define ASCII_CR        0x0D
#define ASCII_XON       0x11
#define ASCII_XOFF      0x13

#define STR_STX         "\x2"
#define STR_ETX         "\x3"


#define _perf_cnt_to_us( _v ) \
         ( (double) (_v) * (double) ( HNS_PER_SEC / 10 ) / (double) PerfFrequency )

#define _perf_cnt_to_hns( _v ) \
         ( (double) (_v) * (double) HNS_PER_SEC / (double) PerfFrequency )

#define _secs_to_hns( _v ) \
         ( ( (LONGLONG) (_v) ) * HNS_PER_SEC )

#if defined( MBG_TGT_WIN32 )

typedef struct
{
  REF_TIME rt;
  MBG_PC_CYCLES cycles;
} REF_TIME_CYCLES;



#define MAX_TIMESTR_LEN 150

typedef struct
{
  char timestring[MAX_TIMESTR_LEN];
  REF_TIME sys_time_on_time;
  MBG_PC_CYCLES sys_cycles_on_time;
} SERIAL_TIME_STRING;


// a pointer to a function that takes a time stamp of the system time
typedef int (*GET_REF_TIME_FNC)( REF_TIME *tp, LARGE_INTEGER *pli );

#endif //MBG_TGT_WIN32

// supported time string formats

enum
{
  SF_MBG_STD,
  SF_SAT,
  SF_INTERFLEX,
  N_SF
};


// Initializers for the time string format names

#define SF_NAME_MBG_STD_ENG  "Meinberg Time String"
#define SF_NAME_MBG_STD_GER  "Meinberg Zeittelegramm"

#define SF_NAME_SAT_ENG      "SAT Time String"
#define SF_NAME_SAT_GER      "SAT Zeittelegramm"

#define SF_NAME_INTERFLEX_ENG "Interflex Time String"
#define SF_NAME_INTERFLEX_GER "Interflex Zeittelegramm"


#define SF_NAMES_ENG     \
{                        \
  SF_NAME_MBG_STD_ENG,   \
  SF_NAME_SAT_ENG,       \
  SF_NAME_INTERFLEX_ENG  \
}


#define SF_NAMES_LSTR                              \
{                                                  \
  { SF_NAME_MBG_STD_ENG, SF_NAME_MBG_STD_GER },    \
  { SF_NAME_SAT_ENG, SF_NAME_SAT_GER },            \
  { SF_NAME_INTERFLEX_ENG, SF_NAME_INTERFLEX_GER } \
}



// a structure to store the characteristics of a time string

typedef struct
{
  const char *chk_fmt;  // to check the format, len must be str len
  char skip_char;
  char first_char;
  char on_time_char;
} TS_SPEC;


#define TS_SPEC_MBG_STD                             \
{                                                   \
  STR_STX "D:__.__.__;T:_;U:__.__.__;____" STR_ETX, \
  '_',                                              \
  ASCII_STX,                                        \
  ASCII_STX                                         \
}

#define TS_SPEC_SAT                              \
{                                                \
  STR_STX "__.__.__/_/__:__:__________" STR_ETX, \
  '_',                                           \
  ASCII_STX,                                     \
  ASCII_STX                                      \
}

#define TS_SPEC_INTERFLEX                             \
{                                                     \
  STR_STX "IFD:__.__.__;T:_;U:__.__.__;____" STR_ETX, \
  '_',                                                \
  ASCII_STX,                                          \
  ASCII_STX                                           \
}


// Default initializer for an array of N_SF TS_SPECs

#define DEFAULT_TS_SPECS \
{                        \
  TS_SPEC_MBG_STD,       \
  TS_SPEC_SAT,           \
  TS_SPEC_INTERFLEX      \
}

#if defined( MBG_TGT_WIN32 )

_ext TS_SPEC ts_spec[N_SF]
#ifdef _DO_INIT
 = DEFAULT_TS_SPECS
#endif
;


_ext ushort year_limit
#ifdef _DO_INIT
 = 1970
#endif
;

#endif //MBG_TGT_WIN32



#if defined( MBG_TGT_WIN32 )

static __mbg_inline
void mbg_pcps_time_stamp_to_filetime( FILETIME *p_ft, const PCPS_TIME_STAMP *p_ts )
{
  uint64_t ts64;
  uint32_t frac;

  frac = bin_frac_32_to_dec_frac( p_ts->frac, PCPS_HRT_FRAC_SCALE );
  ts64 = ( (uint64_t) p_ts->sec ) * HNS_PER_SEC + FILETIME_1970 + frac;

  p_ft->dwHighDateTime = (uint32_t) ( ts64 >> 32 );
  p_ft->dwLowDateTime  = (uint32_t) ( ts64 & 0xFFFFFFFFUL );

}  // mbg_pcps_time_stamp_to_filetime



static __mbg_inline
void mbg_pcps_tstamp64_to_filetime( FILETIME *p_ft, const uint64_t *p_ts64 )
{
  PCPS_TIME_STAMP ts;

  uint64_to_pcps_time_stamp( &ts, *p_ts64 );
  mbg_pcps_time_stamp_to_filetime( p_ft, &ts );

}  // mbg_pcps_tstamp64_to_filetime



/* function prototypes: */

/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 int check_time_string( const char *ts, int len, int sf ) ;
 void decode_str_date_std( SYSTEMTIME *pt, const char *s ) ;
 void decode_str_wday_std( SYSTEMTIME *pt, const char *s ) ;
 void decode_str_time_std( SYSTEMTIME *pt, const char *s ) ;
 int decode_time_str( SYSTEMTIME *pt, long *utc_offs_sec, PCPS_TIME_STATUS_X *status, const char *ts, const MBG_SERCLOCK_CFG *pcfg ) ;
 void timestring_to_ref_and_sys_filetime( SERIAL_TIME_STRING *sts, REF_TIME *st, REF_TIME *rt, const MBG_SERCLOCK_CFG *pcfg ) ;
 void systemtime_to_pcps_time( PCPS_TIME *pt, const SYSTEMTIME *pst ) ;
 void pcps_time_to_systemtime( SYSTEMTIME *pst, const PCPS_TIME *pt ) ;
 int pcps_time_to_ref_time( REF_TIME *rt, const PCPS_TIME *pt ) ;
 void pcps_hr_time_to_ref_time( REF_TIME *rt, const PCPS_HR_TIME *hr_time ) ;
 void pcps_hr_time_to_pcps_time( PCPS_TIME *ptime, const PCPS_HR_TIME *hr_time ) ;

/* ----- function prototypes end ----- */

#endif //MBG_TGT_WIN32

#ifdef __cplusplus
}
#endif


#if defined( _USING_BYTE_ALIGNMENT )
  #pragma pack()      // set default alignment
  #undef _USING_BYTE_ALIGNMENT
#endif

/* End of header body */

#undef _ext
#undef _DO_INIT

#endif  /* _TIMECNV_H */
